diff --git a/common/bs.h b/common/bs.h
index eafa8f8..263de56 100644
--- a/common/bs.h
+++ b/common/bs.h
@@ -169,6 +169,13 @@ static inline void bs_align_1( bs_t *s )
     bs_flush( s );
 }
 
+static inline void bs_align_10( bs_t *s )
+{
+    if( s->i_left&7 ) 
+        bs_write1( s, 1 );
+    bs_flush( s );    
+}
+
 /* golomb functions */
 
 static const uint8_t x264_ue_size_tab[256] =
diff --git a/common/common.c b/common/common.c
index 9260c64..2078142 100644
--- a/common/common.c
+++ b/common/common.c
@@ -72,6 +72,7 @@ void    x264_param_default( x264_param_t *param )
     param->i_bframe_adaptive = X264_B_ADAPT_FAST;
     param->i_bframe_bias = 0;
     param->b_bframe_pyramid = 0;
+    param->b_strictdpb = 0;
 
     param->b_deblocking_filter = 1;
     param->i_deblocking_filter_alphac0 = 0;
@@ -350,6 +351,8 @@ int x264_param_parse( x264_param_t *p, const char *name, const char *value )
         p->i_bframe_bias = atoi(value);
     OPT("b-pyramid")
         p->b_bframe_pyramid = atobool(value);
+    OPT("strictdpb")
+        p->b_strictdpb = atobool(value);
     OPT("nf")
         p->b_deblocking_filter = !atobool(value);
     OPT2("filter", "deblock")
@@ -367,6 +370,8 @@ int x264_param_parse( x264_param_t *p, const char *name, const char *value )
         else
             p->b_deblocking_filter = atobool(value);
     }
+    OPT("open-gop")
+        p->b_open_gop = atobool(value);
     OPT("cabac")
         p->b_cabac = atobool(value);
     OPT("cabac-idc")
@@ -852,6 +857,7 @@ char *x264_param2string( x264_param_t *p, int b_res )
     s += sprintf( s, " cqm=%d", p->i_cqm_preset );
     s += sprintf( s, " deadzone=%d,%d", p->analyse.i_luma_deadzone[0], p->analyse.i_luma_deadzone[1] );
     s += sprintf( s, " chroma_qp_offset=%d", p->analyse.i_chroma_qp_offset );
+    s += sprintf( s, " opengop=%d", p->b_open_gop );
     s += sprintf( s, " threads=%d", p->i_threads );
     s += sprintf( s, " nr=%d", p->analyse.i_noise_reduction );
     s += sprintf( s, " decimate=%d", p->analyse.b_dct_decimate );
@@ -860,8 +866,8 @@ char *x264_param2string( x264_param_t *p, int b_res )
     s += sprintf( s, " bframes=%d", p->i_bframe );
     if( p->i_bframe )
     {
-        s += sprintf( s, " b_pyramid=%d b_adapt=%d b_bias=%d direct=%d wpredb=%d",
-                      p->b_bframe_pyramid, p->i_bframe_adaptive, p->i_bframe_bias,
+        s += sprintf( s, " b_pyramid=%d b_strictdpb=%d b_adapt=%d b_bias=%d direct=%d wpredb=%d",
+                      p->b_bframe_pyramid, p->b_strictdpb, p->i_bframe_adaptive, p->i_bframe_bias,
                       p->analyse.i_direct_mv_pred, p->analyse.b_weighted_bipred );
     }
 
diff --git a/common/common.h b/common/common.h
index 1e46ae8..f146e68 100644
--- a/common/common.h
+++ b/common/common.h
@@ -203,6 +203,11 @@ typedef struct
         int idc;
         int arg;
     } ref_pic_list_order[2][16];
+    struct // for future expansion
+    {
+        int i_difference_of_pic_nums;
+    } mmco[16];
+    int i_mmco_idx;
 
     int i_cabac_init_idc;
 
@@ -338,7 +343,9 @@ struct x264_t
         /* frames used for reference + sentinels */
         x264_frame_t *reference[16+2];
 
-        int i_last_idr; /* Frame number of the last IDR */
+        int i_last_idr; /* Frame number of the last IDR or RP */
+        int i_idr_opengop_offset; /* Offset of above to handle open gop */
+        int b_leading_bframes; /* Place holder to know when to remove old references for soft gop*/
 
         int i_input;    /* Number of input frames already accepted */
 
@@ -363,7 +370,11 @@ struct x264_t
     x264_frame_t    *fref1[16+3];     /* ref list 1 */
     int             b_ref_reorder[2];
 
-
+    int             i_mmco_idx;
+    struct
+    {
+        int         i_difference_of_pic_nums;
+    } mmco[16];
 
     /* Current MB DCT coeffs */
     struct
diff --git a/common/frame.h b/common/frame.h
index aad77f5..6ba2c8f 100644
--- a/common/frame.h
+++ b/common/frame.h
@@ -38,8 +38,11 @@ typedef struct
     int     i_frame;    /* Presentation frame number */
     int     i_frame_num; /* Coded frame number */
     int     b_kept_as_ref;
+    int     b_keyframe;
     float   f_qp_avg_rc; /* QPs as decided by ratecontrol */
     float   f_qp_avg_aq; /* QPs as decided by AQ in addition to ratecontrol */
+    int     i_mmco; /* memory management NULL == no commands*/
+    int     i_mmco_dipnm1[16+2]; /* short-term reference->unused for reference */
 
     /* YUV buffer */
     int     i_plane;
diff --git a/encoder/encoder.c b/encoder/encoder.c
index fdcc957..b256c08 100644
--- a/encoder/encoder.c
+++ b/encoder/encoder.c
@@ -123,6 +123,16 @@ static void x264_slice_header_init( x264_t *h, x264_slice_header_t *sh,
     sh->b_ref_pic_list_reordering_l0 = h->b_ref_reorder[0];
     sh->b_ref_pic_list_reordering_l1 = h->b_ref_reorder[1];
 
+    sh->i_mmco_idx = h->i_mmco_idx;
+    if( h->i_mmco_idx )
+    {
+        int j = 0;
+        for( ; j < h->i_mmco_idx; j++ )
+        {
+            sh->mmco[j].i_difference_of_pic_nums = h->mmco[j].i_difference_of_pic_nums;
+        }
+    }
+
     /* If the ref list isn't in the default order, construct reordering header */
     /* List1 reordering isn't needed yet */
     if( sh->b_ref_pic_list_reordering_l0 )
@@ -175,7 +185,7 @@ static void x264_slice_header_write( bs_t *s, x264_slice_header_t *sh, int i_nal
 
     bs_write_ue( s, sh->i_type + 5 );   /* same type things */
     bs_write_ue( s, sh->i_pps_id );
-    bs_write( s, sh->sps->i_log2_max_frame_num, sh->i_frame_num );
+    bs_write( s, sh->sps->i_log2_max_frame_num, sh->i_frame_num & (1<<sh->sps->i_log2_max_frame_num)-1 );
 
     if( !sh->sps->b_frame_mbs_only )
     {
@@ -191,7 +201,7 @@ static void x264_slice_header_write( bs_t *s, x264_slice_header_t *sh, int i_nal
 
     if( sh->sps->i_poc_type == 0 )
     {
-        bs_write( s, sh->sps->i_log2_max_poc_lsb, sh->i_poc_lsb );
+        bs_write( s, sh->sps->i_log2_max_poc_lsb, sh->i_poc_lsb & (1<<sh->sps->i_log2_max_poc_lsb)-1 );
         if( sh->pps->b_pic_order && !sh->b_field_pic )
         {
             bs_write_se( s, sh->i_delta_poc_bottom );
@@ -272,7 +282,17 @@ static void x264_slice_header_write( bs_t *s, x264_slice_header_t *sh, int i_nal
         }
         else
         {
-            bs_write1( s, 0 );  /* adaptive_ref_pic_marking_mode_flag */
+            bs_write1( s, sh->i_mmco_idx > 0 ); /* adaptive_ref_pic_marking_mode_flag */
+            if( sh->i_mmco_idx > 0 )
+            {
+                int i = 0;
+                for( ; i < sh->i_mmco_idx; i++ )
+                {
+                    bs_write_ue( s, 1 ); /* mark short term ref as unused */
+                    bs_write_ue( s, sh->mmco[i].i_difference_of_pic_nums - 1 );
+                }
+                bs_write_ue( s, 0 );
+            }
         }
     }
 
@@ -733,6 +753,7 @@ x264_t *x264_encoder_open   ( x264_param_t *param )
     h->frames.b_have_sub8x8_esa = !!(h->param.analyse.inter & X264_ANALYSE_PSUB8x8);
 
     h->frames.i_last_idr = - h->param.i_keyint_max;
+    h->frames.i_idr_opengop_offset = 0; //will stay 0 if !open-gop
     h->frames.i_input    = 0;
     h->frames.last_nonb  = NULL;
 
@@ -916,6 +937,13 @@ int x264_encoder_headers( x264_t *h, x264_nal_t **pp_nal, int *pi_nal )
         x264_nal_start( h, NAL_PPS, NAL_PRIORITY_HIGHEST );
         x264_pps_write( &h->out.bs, h->pps );
         x264_nal_end( h );
+        /* recovery point */
+        if( h->param.b_open_gop ) 
+        {
+            x264_nal_start( h, NAL_SEI, NAL_PRIORITY_DISPOSABLE );
+            x264_sei_recovery_point_write( h, &h->out.bs );
+            x264_nal_end( h );
+        }
     }
     /* now set output*/
     *pi_nal = h->out.i_nal;
@@ -986,6 +1014,8 @@ static inline void x264_reference_build_list( x264_t *h, int i_poc )
                 h->b_ref_reorder[0] = 1;
                 break;
             }
+        if( h->param.b_bframe_pyramid )
+            h->b_ref_reorder[0] = 1;
     }
 
     h->i_ref1 = X264_MIN( h->i_ref1, h->frames.i_max_ref1 );
@@ -1115,6 +1145,43 @@ static inline void x264_reference_reset( x264_t *h )
     h->fenc->i_poc = 0;
 }
 
+static inline void x264_reference_partial_reset( x264_t *h, int leave_one_nonbref_only )
+{
+    int refs = 0;
+    int lastnonbidx = -1;
+    while( h->frames.reference[refs] )
+    {
+        if( h->frames.reference[refs]->i_type == X264_TYPE_I
+         || h->frames.reference[refs]->i_type == X264_TYPE_IDR
+       || ( h->frames.reference[refs]->i_type == X264_TYPE_P
+         && !leave_one_nonbref_only) )
+            lastnonbidx = refs;
+        refs++;
+    }
+    assert( lastnonbidx > -1 );
+    while( refs > lastnonbidx+1 )
+    {
+        int diff = h->i_frame_num - h->frames.reference[refs-1]->i_frame_num;
+        h->mmco[ h->i_mmco_idx++ ].i_difference_of_pic_nums = diff;
+        x264_frame_push_unused( h, x264_frame_pop( h->frames.reference ) );
+        refs--;
+    }
+    // This makes room in the dpb for the delayed display time of the later b-frame's
+    // Within a mini-gop . Specific to b-pyramid. Also can remove all but one ref as well depending on 
+    // leave_one_nonbref_only
+    while( ( h->param.b_strictdpb && ( ( refs + 2 ) > X264_MAX( h->frames.i_max_dpb, 4 ) )
+                                  && h->sh.i_type ==  SLICE_TYPE_B
+                                  && !leave_one_nonbref_only )
+                                  || ( lastnonbidx > 0 && leave_one_nonbref_only ) )
+    {
+        int diff = h->i_frame_num - h->frames.reference[0]->i_frame_num;
+        h->mmco[ h->i_mmco_idx++ ].i_difference_of_pic_nums = diff;
+        x264_frame_push_unused( h, x264_frame_shift( h->frames.reference ) );
+        lastnonbidx--;
+        refs--;
+    }
+}
+
 static inline void x264_slice_init( x264_t *h, int i_nal_type, int i_global_qp )
 {
     /* ------------------------ Create slice header  ----------------------- */
@@ -1435,6 +1502,7 @@ int     x264_encoder_encode( x264_t *h,
             x264_frame_expand_border_mod16( h, fenc );
 
         fenc->i_frame = h->frames.i_input++;
+        h->i_mmco_idx = 0;
 
         x264_frame_push( h->frames.next, fenc );
 
@@ -1491,10 +1559,24 @@ int     x264_encoder_encode( x264_t *h,
         pic_out->i_type = X264_TYPE_AUTO;
         return 0;
     }
-
+    int b_exceeds_min = (h->fenc->i_frame - h->frames.i_last_idr ) > h->param.i_keyint_min;
+    h->fdec->b_keyframe = 0;
+    if( h->param.b_open_gop )
+    {
+        if( ( (h->fenc->i_type == X264_TYPE_I && b_exceeds_min) 
+            || h->fenc->i_type == X264_TYPE_IDR ) 
+            && h->fenc->i_frame != 0 )
+        {
+            h->frames.i_idr_opengop_offset += ( h->fenc->i_frame - h->frames.i_last_idr );
+            h->fdec->b_keyframe = 1;
+        }
+        if( h->fenc->i_type == X264_TYPE_I && b_exceeds_min )
+            h->frames.i_last_idr = h->fenc->i_frame;
+    }
     if( h->fenc->i_type == X264_TYPE_IDR )
     {
         h->frames.i_last_idr = h->fenc->i_frame;
+        h->fdec->b_keyframe = 1;
     }
 
     /* ------------------- Setup frame context ----------------------------- */
@@ -1504,6 +1586,7 @@ int     x264_encoder_encode( x264_t *h,
         /* reset ref pictures */
         x264_reference_reset( h );
 
+        h->frames.i_idr_opengop_offset = 0;
         i_nal_type    = NAL_SLICE_IDR;
         i_nal_ref_idc = NAL_PRIORITY_HIGHEST;
         h->sh.i_type = SLICE_TYPE_I;
@@ -1513,18 +1596,29 @@ int     x264_encoder_encode( x264_t *h,
         i_nal_type    = NAL_SLICE;
         i_nal_ref_idc = NAL_PRIORITY_HIGH; /* Not completely true but for now it is (as all I/P are kept as ref)*/
         h->sh.i_type = SLICE_TYPE_I;
+        if( ( h->fdec->b_keyframe && h->frames.b_leading_bframes ) || h->param.b_bframe_pyramid )
+            x264_reference_partial_reset( h, h->frames.b_leading_bframes );
+        if( h->param.b_open_gop )
+        {
+            h->frames.b_leading_bframes = (b_exceeds_min && h->frames.b_leading_bframes) ? 1 : !b_exceeds_min ? 0 : 1;
+        }
     }
     else if( h->fenc->i_type == X264_TYPE_P )
     {
         i_nal_type    = NAL_SLICE;
         i_nal_ref_idc = NAL_PRIORITY_HIGH; /* Not completely true but for now it is (as all I/P are kept as ref)*/
         h->sh.i_type = SLICE_TYPE_P;
+        x264_reference_partial_reset( h, h->frames.b_leading_bframes );
+        if( h->frames.b_leading_bframes ) 
+            h->frames.b_leading_bframes = 0;
     }
     else if( h->fenc->i_type == X264_TYPE_BREF )
     {
         i_nal_type    = NAL_SLICE;
         i_nal_ref_idc = NAL_PRIORITY_HIGH; /* maybe add MMCO to forget it? -> low */
         h->sh.i_type = SLICE_TYPE_B;
+        if( h->param.b_strictdpb )
+            x264_reference_partial_reset( h, 0 );
     }
     else    /* B frame */
     {
@@ -1534,9 +1628,10 @@ int     x264_encoder_encode( x264_t *h,
     }
 
     h->fdec->i_poc =
-    h->fenc->i_poc = 2 * (h->fenc->i_frame - h->frames.i_last_idr);
+    h->fenc->i_poc = 2 * (h->fenc->i_frame - ( h->frames.i_last_idr - h->frames.i_idr_opengop_offset ) );    
     h->fdec->i_type = h->fenc->i_type;
     h->fdec->i_frame = h->fenc->i_frame;
+    h->fdec->i_mmco = h->fenc->i_mmco;
     h->fenc->b_kept_as_ref =
     h->fdec->b_kept_as_ref = i_nal_ref_idc != NAL_PRIORITY_DISPOSABLE && h->param.i_keyint_max > 1;
 
@@ -1589,7 +1684,8 @@ int     x264_encoder_encode( x264_t *h,
     h->i_nal_ref_idc = i_nal_ref_idc;
 
     /* Write SPS and PPS */
-    if( i_nal_type == NAL_SLICE_IDR && h->param.b_repeat_headers )
+    if( (i_nal_type == NAL_SLICE_IDR && h->param.b_repeat_headers)
+                    || ( h->sh.i_type == SLICE_TYPE_I && h->fdec->b_keyframe ) )
     {
         if( h->fenc->i_frame == 0 )
         {
@@ -1608,6 +1704,13 @@ int     x264_encoder_encode( x264_t *h,
         x264_nal_start( h, NAL_PPS, NAL_PRIORITY_HIGHEST );
         x264_pps_write( &h->out.bs, h->pps );
         x264_nal_end( h );
+
+        /* recovery point */
+        if( h->param.b_open_gop ) {
+            x264_nal_start( h, NAL_SEI, NAL_PRIORITY_DISPOSABLE );
+            x264_sei_recovery_point_write( h, &h->out.bs );
+            x264_nal_end( h );
+        }
     }
 
     /* Write frame */
@@ -1663,6 +1766,7 @@ static void x264_encoder_frame_end( x264_t *h, x264_t *thread_current,
         pic_out->img.i_stride[i] = h->fdec->i_stride[i];
         pic_out->img.plane[i] = h->fdec->plane[i];
     }
+    pic_out->b_keyframe = h->fdec->b_keyframe;
 
     /* ---------------------- Update encoder state ------------------------- */
 
diff --git a/encoder/set.c b/encoder/set.c
index 3103fcd..d4ba1bb 100644
--- a/encoder/set.c
+++ b/encoder/set.c
@@ -77,6 +77,7 @@ static void scaling_list_write( bs_t *s, x264_pps_t *pps, int idx )
 void x264_sps_init( x264_sps_t *sps, int i_id, x264_param_t *param )
 {
     sps->i_id = i_id;
+    int max_frame_num;
 
     sps->b_qpprime_y_zero_transform_bypass = param->rc.i_rc_method == X264_RC_CQP && param->rc.i_qp_constant == 0;
     if( sps->b_qpprime_y_zero_transform_bypass )
@@ -96,17 +97,26 @@ void x264_sps_init( x264_sps_t *sps, int i_id, x264_param_t *param )
     /* Never set constraint_set2, it is not necessary and not used in real world. */
     sps->b_constraint_set2  = 0;
 
-    sps->i_log2_max_frame_num = 4;  /* at least 4 */
-    while( (1 << sps->i_log2_max_frame_num) <= param->i_keyint_max )
-    {
+    sps->vui.i_num_reorder_frames = param->b_bframe_pyramid ? 2 : param->i_bframe ? 1 : 0;
+    /* extra slot with pyramid so that we don't have to override the
+     * order of forgetting old pictures */
+    sps->vui.i_max_dec_frame_buffering =
+    sps->i_num_ref_frames = X264_MIN(16, X264_MAX(param->i_frame_reference, 1 + sps->vui.i_num_reorder_frames));
+
+    /* number of refs + current frame */
+    /* we MMCO away b-refs, so we have to count those too. */
+    max_frame_num = sps->i_num_ref_frames * (param->b_bframe_pyramid?2:1) + 1;
+    sps->i_log2_max_frame_num = 4;
+    while( (1 << sps->i_log2_max_frame_num) < max_frame_num )
         sps->i_log2_max_frame_num++;
-    }
-    sps->i_log2_max_frame_num++;    /* just in case */
 
     sps->i_poc_type = 0;
     if( sps->i_poc_type == 0 )
     {
-        sps->i_log2_max_poc_lsb = sps->i_log2_max_frame_num + 1;    /* max poc = 2*frame_num */
+        int max_poc = (param->i_bframe+1)*(param->b_bframe_pyramid?3:2)*2;
+        sps->i_log2_max_poc_lsb = 4;
+        while( (1 << sps->i_log2_max_poc_lsb) < max_poc )
+            sps->i_log2_max_poc_lsb++;
     }
     else if( sps->i_poc_type == 1 )
     {
@@ -193,12 +203,6 @@ void x264_sps_init( x264_sps_t *sps, int i_id, x264_param_t *param )
         sps->vui.b_fixed_frame_rate = 1;
     }
 
-    sps->vui.i_num_reorder_frames = param->b_bframe_pyramid ? 2 : param->i_bframe ? 1 : 0;
-    /* extra slot with pyramid so that we don't have to override the
-     * order of forgetting old pictures */
-    sps->vui.i_max_dec_frame_buffering =
-    sps->i_num_ref_frames = X264_MIN(16, X264_MAX(param->i_frame_reference, 1 + sps->vui.i_num_reorder_frames));
-
     sps->vui.b_bitstream_restriction = 1;
     if( sps->vui.b_bitstream_restriction )
     {
@@ -470,6 +474,22 @@ void x264_pps_write( bs_t *s, x264_pps_t *pps )
     bs_rbsp_trailing( s );
 }
 
+void x264_sei_recovery_point_write( x264_t *h, bs_t *s )
+{
+    int payload_size;
+
+    bs_write( s, 8, 0x06 ); // payload_type = Recovery Point
+    payload_size = 1 ;  // values as given below will never be > 1    
+
+    bs_write( s, 8, (payload_size + 7) / 8);
+    bs_write_ue( s, 0 ); // recovery_frame_cnt 0
+    bs_write( s, 1, 1 ); //exact_match_flag 1
+    bs_write( s, 1, 0 ); //broken_link_flag 0
+    bs_write( s, 2, 0 ); //changing_slice_group 0    
+
+    bs_align_10( s );
+}
+
 void x264_sei_version_write( x264_t *h, bs_t *s )
 {
     int i;
diff --git a/encoder/set.h b/encoder/set.h
index 3611c9a..cb1c076 100644
--- a/encoder/set.h
+++ b/encoder/set.h
@@ -24,6 +24,7 @@
 #ifndef X264_ENCODER_SET_H
 #define X264_ENCODER_SET_H
 
+void x264_sei_recovery_point_write( x264_t *h, bs_t *s );
 void x264_sps_init( x264_sps_t *sps, int i_id, x264_param_t *param );
 void x264_sps_write( bs_t *s, x264_sps_t *sps );
 void x264_pps_init( x264_pps_t *pps, int i_id, x264_param_t *param, x264_sps_t *sps );
diff --git a/encoder/slicetype.c b/encoder/slicetype.c
index 2c16429..c32dbf5 100644
--- a/encoder/slicetype.c
+++ b/encoder/slicetype.c
@@ -487,6 +487,7 @@ static void x264_slicetype_analyse( x264_t *h )
     int i_mb_count = NUM_MBS;
     int cost1p0, cost2p0, cost1b1, cost2p1;
     int idr_frame_type;
+    int minigop_end_frame_type[h->param.i_bframe+2];
 
     assert( h->frames.b_have_lowres );
 
@@ -495,18 +496,21 @@ static void x264_slicetype_analyse( x264_t *h )
     frames[0] = h->frames.last_nonb;
     for( j = 0; h->frames.next[j] && h->frames.next[j]->i_type == X264_TYPE_AUTO; j++ )
         frames[j+1] = h->frames.next[j];
-    keyint_limit = h->param.i_keyint_max - frames[0]->i_frame + h->frames.i_last_idr - 1;
+    keyint_limit = h->param.i_keyint_max - frames[0]->i_frame + h->frames.i_last_idr;
     num_frames = X264_MIN( j, keyint_limit );
     if( num_frames == 0 )
         return;
 
     x264_lowres_context_init( h, &a );
-    idr_frame_type = frames[1]->i_frame - h->frames.i_last_idr >= h->param.i_keyint_min ? X264_TYPE_IDR : X264_TYPE_I;
-
+    idr_frame_type = frames[1]->i_frame - h->frames.i_last_idr >= h->param.i_keyint_min ?
+                 ( h->param.b_open_gop && h->frames.i_last_idr > -1 ? X264_TYPE_I : X264_TYPE_IDR ) : X264_TYPE_I;
+    for( j = 1; j < h->param.i_bframe+2 && frames[j]; j++ )
+        minigop_end_frame_type[j] = frames[j]->i_frame - h->frames.i_last_idr
+                                 >= h->param.i_keyint_max ? X264_TYPE_I : X264_TYPE_P;
     if( num_frames == 1 )
     {
 no_b_frames:
-        frames[1]->i_type = X264_TYPE_P;
+        frames[1]->i_type = minigop_end_frame_type[1];
         if( h->param.i_scenecut_threshold && scenecut( h, &a, frames, 0, 1 ) )
             frames[1]->i_type = idr_frame_type;
         return;
@@ -533,7 +537,7 @@ no_b_frames:
             }
             frames[j]->i_type = X264_TYPE_B;
         }
-        frames[num_bframes+1]->i_type = X264_TYPE_P;
+        frames[num_bframes+1]->i_type = minigop_end_frame_type[j];
     }
     else if( h->param.i_bframe_adaptive == X264_B_ADAPT_FAST )
     {
@@ -566,8 +570,9 @@ no_b_frames:
             else
                 frames[j]->i_type = X264_TYPE_B;
         }
+        frames[j]->i_type = minigop_end_frame_type[j];
     }
-    else
+    else // X264_B_ADAPT_NONE
     {
         int max_bframes = X264_MIN(num_frames-1, h->param.i_bframe);
         if( h->param.i_scenecut_threshold && scenecut( h, &a, frames, 0, 1 ) )
@@ -585,7 +590,7 @@ no_b_frames:
             }
             frames[j]->i_type = X264_TYPE_B;
         }
-        frames[max_bframes+1]->i_type = X264_TYPE_P;
+        frames[max_bframes+1]->i_type = minigop_end_frame_type[j];
     }
 }
 
@@ -612,17 +617,19 @@ void x264_slicetype_decide( x264_t *h )
     for( bframes = 0;; bframes++ )
     {
         frm = h->frames.next[bframes];
-
         /* Limit GOP size */
         if( frm->i_frame - h->frames.i_last_idr >= h->param.i_keyint_max )
         {
             if( frm->i_type == X264_TYPE_AUTO )
-                frm->i_type = X264_TYPE_IDR;
-            if( frm->i_type != X264_TYPE_IDR )
+                frm->i_type = h->param.b_open_gop && h->frames.i_last_idr > -1 ? X264_TYPE_I : X264_TYPE_IDR;
+            if( frm->i_type != X264_TYPE_I && frm->i_type != X264_TYPE_IDR )
                 x264_log( h, X264_LOG_WARNING, "specified frame type (%d) is not compatible with keyframe interval\n", frm->i_type );
         }
-        if( frm->i_type == X264_TYPE_IDR )
+        if( ( frm->i_type == X264_TYPE_I || frm->i_type == X264_TYPE_IDR )
+                          && frm->i_frame - h->frames.i_last_idr >= h->param.i_keyint_max
+                          && !h->param.b_open_gop )
         {
+            frm->i_type = X264_TYPE_IDR;
             /* Close GOP */
             if( bframes > 0 )
             {
@@ -634,7 +641,6 @@ void x264_slicetype_decide( x264_t *h )
                 h->i_frame_num = 0;
             }
         }
-
         if( bframes == h->param.i_bframe
             || h->frames.next[bframes+1] == NULL )
         {
diff --git a/muxers.c b/muxers.c
index d62be5c..d924039 100644
--- a/muxers.c
+++ b/muxers.c
@@ -780,7 +780,7 @@ int set_eop_mp4( hnd_t handle, x264_picture_t *p_picture )
     uint64_t pts = (uint64_t)p_picture->i_pts;
     int32_t offset = p_mp4->i_init_delay + pts - dts;
 
-    p_mp4->p_sample->IsRAP = p_picture->i_type == X264_TYPE_IDR ? 1 : 0;
+    p_mp4->p_sample->IsRAP = p_picture->b_keyframe;
     p_mp4->p_sample->DTS = dts;
     p_mp4->p_sample->CTS_Offset = offset;
     gf_isom_add_sample(p_mp4->p_file, p_mp4->i_track, p_mp4->i_descidx, p_mp4->p_sample);
@@ -996,8 +996,7 @@ int set_eop_mkv( hnd_t handle, x264_picture_t *p_picture )
 
     p_mkv->b_writing_frame = 0;
 
-    return mk_setFrameFlags( p_mkv->w, i_stamp,
-                             p_picture->i_type == X264_TYPE_IDR );
+    return mk_setFrameFlags( p_mkv->w, i_stamp, p_picture->b_keyframe );
 }
 
 int close_file_mkv( hnd_t handle )
diff --git a/x264.c b/x264.c
index 91d2ca1..e329b2a 100644
--- a/x264.c
+++ b/x264.c
@@ -191,6 +191,8 @@ static void Help( x264_param_t *defaults, int b_longhelp )
         "                                  - 2: Optimal (slow with high --bframes)\n", defaults->i_bframe_adaptive );
     H1( "      --b-bias <integer>      Influences how often B-frames are used [%d]\n", defaults->i_bframe_bias );
     H0( "      --b-pyramid             Keep some B-frames as references\n" );
+    H0( "      --strictdpb             Stricter usage of dpb when b-pyramid is active (ref 4+)\n" );
+    H0( "      --open-gop              Use non IDR i-frames at max-keyint w/ recovery pt SEI\n" );
     H0( "      --no-cabac              Disable CABAC\n" );
     H0( "  -r, --ref <integer>         Number of reference frames [%d]\n", defaults->i_frame_reference );
     H1( "      --no-deblock            Disable loop filter\n" );
@@ -384,9 +386,11 @@ static struct option long_options[] =
     { "no-b-adapt",        no_argument, NULL, 0 },
     { "b-bias",      required_argument, NULL, 0 },
     { "b-pyramid",         no_argument, NULL, 0 },
+    { "strictdpb",         no_argument, NULL, 0 },
     { "min-keyint",  required_argument, NULL, 'i' },
     { "keyint",      required_argument, NULL, 'I' },
     { "scenecut",    required_argument, NULL, 0 },
+    { "open-gop",          no_argument, NULL, 0 },
     { "no-scenecut",       no_argument, NULL, 0 },
     { "nf",                no_argument, NULL, 0 },
     { "no-deblock",        no_argument, NULL, 0 },
diff --git a/x264.h b/x264.h
index e61040e..168e3ff 100644
--- a/x264.h
+++ b/x264.h
@@ -191,6 +191,8 @@ typedef struct x264_param_t
     int         i_bframe_adaptive;
     int         i_bframe_bias;
     int         b_bframe_pyramid;   /* Keep some B-frames as references */
+    int         b_strictdpb;        /* Reserve a spot in dpb for delayed display time b frames */
+    int         b_open_gop;
 
     int         b_deblocking_filter;
     int         i_deblocking_filter_alphac0;    /* [-6, 6] -6 light filter, 6 strong */
@@ -350,6 +352,7 @@ typedef struct
     int     i_qpplus1;
     /* In: user pts, Out: pts of encoded picture (user)*/
     int64_t i_pts;
+    int     b_keyframe;
 
     /* In: raw data */
     x264_image_t img;